;; Makro co-dotimes
;; Syntax:
;;   (co-dotimes (var form) . body)
;;
;; Funguje podobně jako makro dotimes, ale každou iteraci vyhodnocuje paralelně
;; v různých procesech. Čeká, až všechny procesy skončí.

(defmethod compile-op ((op (eql 'co-dotimes)) args)
  (let ((var (caar args))
        (count-form (cadar args))
        (body (cdr args))
        (fun (gensym "FUN"))
        ;; Semaforem zařídíme čekání na skončení procesů:
        (sem (gensym "SEM"))
        (count (gensym "COUNT")))
    (sub 
     `(let ((,sem (sem 0))
            ;; Funkce, která se bude volat paralelně:
            (,fun (lambda (,var) (begin ,@body (v ,sem))))
            (,count ,count-form))
        ;; Spuštění procesů:
        (dotimes (i ,count)
          (process-run-function ,fun i))
        ;; Čekání, až procesy skončí:
        (dotimes (i ,count)
          (p ,sem))))))

;; Nastavení odsazení operátoru co-dotimes
(editor:setup-indent "co-dotimes" 1)

#|
;; Scheme:
;; Paralelně desetkrát vytiskne čísla od nuly po deset:
(co-dotimes (i 10)
  (dotimes (j 10)
    (print i)))


;; V deseti procesech stokrát inkrementujeme počítadlo:
(let ((n 0)
      (l (lock)))
  (co-dotimes (i 10)
    (dotimes (j 100)
      (with-lock (l)
        (set! n (+ n 1)))))
  n)
|#
