;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; 05_semaphore.lisp - Semafory
;;;

;;
;; Optimalizace operátoru dotimes
;;

;; dotimes přestane být makro
(evaluate '(set! macros (remove 'dotimes macros)))

;; dotimes začne být speciálním operátorem
(evaluate '(set! specials (cons 'dotimes specials)))

;; Procedura vykonávající dotimes.
;; Vstup: ((i n-expr) b1 ... bn)
(execute '(:split begin :swap :cons        ; (begin b1 ... bn) (i n-expr)
           :swap :split :split :drop       ; n-expr i (begin b1 ... bn)
           eval :val :clsub                ; n i (begin b1 ... bn)
           :swap :dup 0 :swap :bind        ; i n (begin b1 ... bn)
           :lbl
               :dup :val :rot           ; n i-val i (begin b1 ... bn)
               :dup :rot-               ; n i-val n i (begin b1 ... bn) 
               :<                       ; nil/t n i  (begin b1 ... bn)
               :if 
                   nil
               :else 
                   :rot                        ; (begin b1 ... bn) n i
                   :dup eval :val :clsub :drop
                   :rot-                       ; n i (begin b1 ... bn)
                   :swap                       ; i n (begin b1 ... bn)
                   :dup :val 1 :+              ; i-val+1 i n (begin b1 ... bn)
                   :over :set! :drop           ; i n (begin b1 ... bn)
                   t
               :then
           :cjmp
           :drop :drop :drop nil :ret) 'dotimes :set!)

;;
;; Semafory
;;

#|
Semafor je reprezentován seznamem:

(value . blocked)

kde value je hodnota semaforu (nezáporné celé číslo) a blocked je seznam procesů zablokovaných semaforem.

Semafor bude férový: blokované procesy ukládá do fronty
|#

;; Vytvoření semaforu:
;; (val -- semafor s hodnotou val)
(defprim :sem 
  (let ((val (pop *rslt*)))
    (unless (<= 0 val)
      (error "Value can not be negative"))
    (push (cons val nil) *rslt*)))

;; Operace počkat
(defprim :p 
  (let ((sem (pop *rslt*)))
    (cond 
     ((plusp (car sem))
      (decf (car sem)))
     (t
      (setf (cdr sem) 
            (append (cdr sem) 
                    (list (running-process))))
      (dispatcher)))))

;; Operace vyhlásit
(defprim :v
  (let ((sem (pop *rslt*)))
    (if (cdr sem)
        (insert-ready-process (pop (cdr sem)))
      (incf (car sem)))))


#|
;; Stacks:
0 :sem
:dup
(1 :print :drop :v :quit) :prsub
:p

;; Vysvětlete:
0 :sem :p

0 :sem s :bind
nil (1 :print s :val :v :quit) :prsub
s :val :p
:unbind

|#

;; Procedury na práci se semafory:
(execute '(:sem :ret) 'sem :bind)
(execute '(:p nil :ret) 'p :bind)
(execute '(:v nil :ret) 'v :bind)
#|
;; Scheme:
(let ((s (sem 0)))
  (co
   (begin
     (print 1)
     (v s))
   (begin
     (p s)
     (print 2))))

;; Neošetřená kritická sekce:
(let ((n 0))
  (co
   (dotimes (i 100)
     (set! n (+ n 1)))
   (dotimes (i 100)
     (set! n (+ n 1))))
  n)


;; Použití semaforu k ošetření kritické sekce:
(let ((n 0)
      (s (sem 1)))
  (co
   (dotimes (i 100)
     (p s)
     (set! n (+ n 1))
     (v s))
   (dotimes (i 100)
     (p s)
     (set! n (+ n 1)) 
     (v s)))
  n)
|#

;;
;; Přepsání makra co
;;

;; Marko co změníme tak, aby čekání
;; na ukončení vytvořených procesů používalo semafory.

;; Procedura počítající délku seznamu
(execute '(:dup :if :drop 0 :else :split :swap :drop length :val :clsub 1  :+ :then :ret)
         'length :bind)


;; Makro co už nemusí být rekurzivní
;; Navíc bude používat semafory k synchronizaci.
(evaluate '(set! co (lambda (args)
                      (let ((start (gensym "START"))
                            (end (gensym "END"))
                            (arg (gensym "ARG"))
                            (args-count (length args)))
                        (cons 'let (cons (list (list start (list 'sem 0))
                                               (list end (list 'sem 0)))
                                         (append
                                          (mapcar (lambda (arg)
                                                    (list 'process-run-procedure
                                                          (list 'lambda (list arg)
                                                                (list 'p start)
                                                                arg
                                                                (list 'v end))
                                                          nil))
                                                  args)
                                          (list
                                           (list 'dotimes (list 'i args-count)
                                                 (list 'v start))
                                           (list 'dotimes (list 'i args-count)
                                                 (list 'p end))))))))))
                        

;; Pomocné makro xco už nebudeme potřebovat:
(evaluate '(set! macros (remove 'xco macros)))
#|
Scheme:

(begin
  (co
   (dotimes (i 10)
     (print 1))
   (dotimes (i 10)
     (print 2)))
  3)

;; Vysvětlete:
(let ((s (sem 0)))
  (co
   (begin
     (p s)
     (print 1)))
  2)
|#

;;
;; Zámky
;;

;; Procedura na vytvoření zámku:
(evaluate '(define lock (lambda ()
                          (sem 1))))


;; Operátor with-lock
;;
;; Syntax: (with-lock (lock) . body)
;;
;; Vyhodnotí tělo za použití zámku lock. 
;; Vrátí hodnotu posledního výrazu těla.

#|
;; with-lock jako makro
(evaluate '(defmac with-lock (lambda (args)
                               (let ((lock-expr (car (car args)))
                                     (body (cdr args))
                                     (lock-sym (gensym "LOCK")))
                                 (list 'let (list (list lock-sym lock-expr))
                                    (list 'p lock-sym)
                                    (list 'let (list (list 'result (cons 'begin body)))
                                      (list 'v lock-sym)
                                      'result))))))
                               
|#

;; with-lock jako speciální operátor
(evaluate '(set! specials (cons 'with-lock specials)))
;; vstup: ((lock) body1 ... exprn)
(execute '(:split :swap :split :drop ; lock (body1 ... exprn)
           eval :val :clsub          ; lock-val lock-val (body1 ... exprn)
           :dup :p                   ; lock-val (body1 ... exprn)
           :swap begin :swap :cons   ; (begin body1 ... exprn) lock-val
           eval :val :clsub          ; body-val lock-val
           :swap :v                  ; body-val
           :ret) 'with-lock :bind)

#|
;; Scheme

;; Použití zámku k ošetření kritické sekce:
(let ((n 0)
      (l (lock)))
  (co
   (dotimes (i 50)
     (with-lock (l)
       (set! n (+ n 1))))
   (dotimes (i 50)
     (with-lock (l)
       (set! n (+ n 1)))))
  n)

;; To samé s tiskem:
(let ((n 0)
      (l (lock)))
  (co
   (dotimes (i 50)
     (with-lock (l)
       (set! n (+ n 1))
       (print (list 1 i n))))
   (dotimes (i 50)
     (with-lock (l)
       (set! n (+ n 1))
       (print (list 2 i n)))))
  n)
|#
